# -*- coding:utf-8 -*-

import warnings

from .table import hanja_table


def translate_syllable(previous, current):
    """Translates a single syllable."""
    from .hangul import dooeum

    if current in hanja_table:
        return dooeum(previous, hanja_table[current])

    return current


def split_hanja(text):
    """주어진 문장을 한자로 된 구역과 그 이외의 문자로 된 구역으로 분리"""

    # TODO: Can we make this a bit prettier?
    if len(text) == 0:
        yield text
    else:
        ch = text[0]
        bucket = [ch]
        prev_state = is_hanja(ch)

        for ch in text[1:]:
            state = is_hanja(ch)

            if prev_state != state:
                yield "".join(bucket)
                bucket = [ch]
            else:
                bucket.append(ch)

            prev_state = state

        yield "".join(bucket)


def is_valid_mode(mode):
    if mode in (
        "substitution",
        "combination-text",
        "combination-text-reversed",
        "combination-html",
    ):
        return True
    elif mode == "combination":
        warnings.warn(
            "Translation mode 'combination' has been deprecated since 0.13.0. "
            "Use 'combination-html' instead."
        )
        return True
    else:
        return False


def get_format_string(mode, word):
    """
    :param mode: substitution | combination-text | combination-text-reversed | combination-html
    """
    if not is_valid_mode(mode):
        raise ValueError(f"Unsupported translation mode: {mode}")

    if mode == "combination-text" and is_hanja(word[0]):
        return "{word}({translated})"
    elif mode == "combination-text-reversed" and is_hanja(word[0]):
        return "{translated}({word})"
    elif mode in ("combination-html", "combination") and is_hanja(word[0]):
        return '<span class="hanja">{word}</span><span class="hangul">({translated})</span>'
    else:
        return "{translated}"


def translate(text, mode):
    """Translates entire text."""
    words = list(split_hanja(text))
    return "".join(
        map(
            lambda w, prev: translate_word(w, prev, get_format_string(mode, w)),
            words,
            [None] + words[:-1],
        )
    )


def translate_word(word, prev, format_string):
    """Translates a single word.

    :param word: Word to be translated
    :param prev: Preceeding word
    """
    prev_char = prev[-1] if prev else " "
    buf = []
    for c in word:
        new_char = translate_syllable(prev_char, c)
        buf.append(new_char)
        prev_char = new_char
    translated = "".join(buf)

    return format_string.format(word=word, translated=translated)


def is_hanja(ch):
    """Determines if a given character ``ch`` is a Chinese character."""
    return ch in hanja_table



# -*- encoding:utf8 -*-
import re
from typing import Dict, List, Tuple, Union

import numpy as np

kor_begin: int = 44032
kor_end: int = 55203
chosung_base: int = 588
jungsung_base: int = 28
jaum_begin: int = 12593
jaum_end: int = 12622
moum_begin: int = 12623
moum_end: int = 12643

chosung_list = [
    "ㄱ",
    "ㄲ",
    "ㄴ",
    "ㄷ",
    "ㄸ",
    "ㄹ",
    "ㅁ",
    "ㅂ",
    "ㅃ",
    "ㅅ",
    "ㅆ",
    "ㅇ",
    "ㅈ",
    "ㅉ",
    "ㅊ",
    "ㅋ",
    "ㅌ",
    "ㅍ",
    "ㅎ",
]

jungsung_list = [
    "ㅏ",
    "ㅐ",
    "ㅑ",
    "ㅒ",
    "ㅓ",
    "ㅔ",
    "ㅕ",
    "ㅖ",
    "ㅗ",
    "ㅘ",
    "ㅙ",
    "ㅚ",
    "ㅛ",
    "ㅜ",
    "ㅝ",
    "ㅞ",
    "ㅟ",
    "ㅠ",
    "ㅡ",
    "ㅢ",
    "ㅣ",
]

jongsung_list = [
    " ",
    "ㄱ",
    "ㄲ",
    "ㄳ",
    "ㄴ",
    "ㄵ",
    "ㄶ",
    "ㄷ",
    "ㄹ",
    "ㄺ",
    "ㄻ",
    "ㄼ",
    "ㄽ",
    "ㄾ",
    "ㄿ",
    "ㅀ",
    "ㅁ",
    "ㅂ",
    "ㅄ",
    "ㅅ",
    "ㅆ",
    "ㅇ",
    "ㅈ",
    "ㅊ",
    "ㅋ",
    "ㅌ",
    "ㅍ",
    "ㅎ",
]

jaum_list = [
    "ㄱ",
    "ㄲ",
    "ㄳ",
    "ㄴ",
    "ㄵ",
    "ㄶ",
    "ㄷ",
    "ㄸ",
    "ㄹ",
    "ㄺ",
    "ㄻ",
    "ㄼ",
    "ㄽ",
    "ㄾ",
    "ㄿ",
    "ㅀ",
    "ㅁ",
    "ㅂ",
    "ㅃ",
    "ㅄ",
    "ㅅ",
    "ㅆ",
    "ㅇ",
    "ㅈ",
    "ㅉ",
    "ㅊ",
    "ㅋ",
    "ㅌ",
    "ㅍ",
    "ㅎ",
]

moum_list = [
    "ㅏ",
    "ㅐ",
    "ㅑ",
    "ㅒ",
    "ㅓ",
    "ㅔ",
    "ㅕ",
    "ㅖ",
    "ㅗ",
    "ㅘ",
    "ㅙ",
    "ㅚ",
    "ㅛ",
    "ㅜ",
    "ㅝ",
    "ㅞ",
    "ㅟ",
    "ㅠ",
    "ㅡ",
    "ㅢ",
    "ㅣ",
]

doublespace_pattern = re.compile(r"\s+")
repeatchars_pattern = re.compile(r"(\w)\\1{3,}")


class ConvolutionHangleEncoder:
    """초/중/종성을 구성하는 자음/모음과 띄어쓰기만 인코딩
    one hot vector [ㄱ, ㄴ, ㄷ, ... ㅎ, ㅏ, ㅐ, .. ㅢ, ㅣ,"  ", ㄱ, ㄲ, ... ㅍ, ㅎ,"  ", 0, 1, 2, .. 9]
    """

    def __init__(self):
        self.jung_begin: int = 19  # len(chosung_list)
        self.jong_begin: int = 40  # self.jung_begin + len(jungsung_list)
        self.number_begin: int = 68  # self.jong_begin + len(jongsung_list)
        self.space: int = (
            78  # len(chosung_list) + len(jungsung_list) + len(jongsung_list) + 10
        )
        self.unk: int = 79
        self.dim: int = 80
        num = [str(i) for i in range(10)]
        space = " "
        unk = "<unk>"
        idx_to_char = (
            chosung_list + jungsung_list + jongsung_list + num + [space] + [unk]
        )
        self.idx_to_char: Dict[int, str] = dict(enumerate(idx_to_char))
        self.jamo_to_idx: Dict[str, int] = {
            "ㄱ": 0,
            "ㄲ": 1,
            "ㄴ": 2,
            "ㄷ": 3,
            "ㄸ": 4,
            "ㄹ": 5,
            "ㅁ": 6,
            "ㅂ": 7,
            "ㅃ": 8,
            "ㅅ": 9,
            "ㅆ": 10,
            "ㅇ": 11,
            "ㅈ": 12,
            "ㅉ": 13,
            "ㅊ": 14,
            "ㅋ": 15,
            "ㅌ": 16,
            "ㅍ": 17,
            "ㅎ": 18,
            "ㅏ": 19,
            "ㅐ": 20,
            "ㅑ": 21,
            "ㅒ": 22,
            "ㅓ": 23,
            "ㅔ": 24,
            "ㅕ": 25,
            "ㅖ": 26,
            "ㅗ": 27,
            "ㅘ": 28,
            "ㅙ": 29,
            "ㅚ": 30,
            "ㅛ": 31,
            "ㅜ": 32,
            "ㅝ": 33,
            "ㅞ": 34,
            "ㅟ": 35,
            "ㅠ": 36,
            "ㅡ": 37,
            "ㅢ": 38,
            "ㅣ": 39,
            " ": 40,
            "ㄳ": 43,
            "ㄵ": 45,
            "ㄶ": 46,
            "ㄺ": 49,
            "ㄻ": 50,
            "ㄼ": 51,
            "ㄽ": 52,
            "ㄾ": 53,
            "ㄿ": 54,
            "ㅀ": 55,
            "ㅄ": 58,
        }

    def encode(self, sent: str) -> np.ndarray:
        onehot = self.sent_to_onehot(sent)
        x = np.zeros((len(onehot), self.dim))
        for i, xi in enumerate(onehot):
            for j in xi:
                if j >= 0:
                    x[i, j] = 1
        return x

    def sent_to_onehot(self, sent: str) -> List[Tuple[int, int, int]]:
        chars = self._normalize(sent)
        ords = [ord(c) for c in chars]
        onehot = []
        for char, idx in zip(chars, ords):
            if idx == 32:
                onehot.append((self.space, -1, -1))
            elif 48 <= idx <= 57:
                onehot.append((idx - 48 + self.number_begin, -1, -1))
            else:
                onehot.append(self._decompose(char, idx))
        return onehot

    def onehot_to_sent(self, encoded_sent: List[Tuple[int, int, int]]) -> str:
        def check_cjj(char: Tuple[int, int, int]):
            cho, jung, jong = char
            if not (0 <= cho < self.jung_begin):
                raise ValueError("Chosung %d is out of index" % cho)
            if not (self.jung_begin <= jung < self.jong_begin):
                raise ValueError("Jungsung %d is out of index" % jung)
            if not (self.jong_begin <= jong < self.number_begin):
                raise ValueError("Jongsung %d is out of index" % jong)

        chars = []
        for char in encoded_sent:
            if char[1] < 0 or char[2] < 0:
                if not 0 <= char[0] < self.dim:
                    raise ValueError(
                        "character index %d is out of index [0, %d]"
                        % (char[0], self.dim)
                    )
                chars.append(self.idx_to_char[char[0]])
            else:
                check_cjj(char)
                cho, jung, jong = tuple(self.idx_to_char[ci] for ci in char)
                chars.append(compose(cho, jung, jong))
        return "".join(chars)

    def _normalize(self, sent: str) -> str:
        import re

        regex = re.compile("[^ㄱ-ㅎㅏ-ㅣ가-힣 0-9]")
        sent = regex.sub(" ", sent)
        sent = doublespace_pattern.sub(" ", sent).strip()
        return sent

    def _compose(
        self,
        cho: int,
        jung: int,
        jong: int,
    ) -> str:
        return chr(kor_begin + chosung_base * cho + jungsung_base * jung + jong)

    def _decompose(self, char: str, i: int) -> Tuple[int, int, int]:
        if not kor_begin <= i <= kor_end:
            return (self.jamo_to_idx.get(char, self.unk), -1, -1)
        i -= kor_begin
        cho = i // chosung_base
        jung = (i - cho * chosung_base) // jungsung_base
        jong = i - cho * chosung_base - jung * jungsung_base
        return (cho, self.jung_begin + jung, self.jong_begin + jong)


def compose(chosung: str, jungsung: str, jongsung: str) -> str:
    return chr(
        kor_begin
        + chosung_base * chosung_list.index(chosung)
        + jungsung_base * jungsung_list.index(jungsung)
        + jongsung_list.index(jongsung)
    )


def decompose(char: str) -> Tuple[str, str, str]:
    if not is_korean(char):
        return (char, " ", " ")
    i = to_base(char)
    if jaum_begin <= i <= jaum_end:
        return (char, " ", " ")
    if moum_begin <= i <= moum_end:
        return (" ", char, " ")
    i -= kor_begin
    cho = i // chosung_base
    jung = (i - cho * chosung_base) // jungsung_base
    jong = i - cho * chosung_base - jung * jungsung_base
    return (chosung_list[cho], jungsung_list[jung], jongsung_list[jong])


def is_korean(char: str):
    i = to_base(char)
    return (
        (kor_begin <= i <= kor_end)
        or (jaum_begin <= i <= jaum_end)
        or (moum_begin <= i <= moum_end)
    )


def is_complete_korean(char: str):
    return kor_begin <= to_base(char) <= kor_end


def is_jaum(char: str):
    return jaum_begin <= to_base(char) <= jaum_end


def is_moum(char: str):
    return moum_begin <= to_base(char) <= moum_end


def to_base(char: Union[str, int]):
    if type(char) in [str, int]:
        return ord(str(char))
    else:
        raise TypeError


def is_number(char: str):
    i = to_base(char)
    return i >= 48 and i <= 57


def is_english(char: str):
    i = to_base(char)
    return (i >= 97 and i <= 122) or (i >= 65 and i <= 90)


def is_punctuation(char: str):
    i = to_base(char)
    return i in [33, 34, 39, 44, 46, 63, 96]



import os

import yaml


def load_table(filename):
    """Loads the Hanja table."""
    try:
        from yaml import CLoader as Loader
    except ImportError:
        from yaml import BaseLoader as Loader
    with open(filename) as fin:
        table = yaml.load(fin.read(), Loader=Loader)

    return table


basepath = os.path.abspath(os.path.dirname(__file__))
hanja_table = load_table(os.path.join(basepath, "table.yml"))



# -*- coding:utf-8 -*-
"""두음법칙에 관련된 내용은
http://ko.wikipedia.org/wiki/%EB%91%90%EC%9D%8C_%EB%B2%95%EC%B9%99 를 참고.
"""
from .impl import translate

__author__ = "Sumin Byeon"
__email__ = "suminb@gmail.com"
__version__ = "0.13.3"



from typing import Callable, List, Optional, Union

from pydantic import BaseModel, model_validator

from lexikanon import HyFI

logger = HyFI.getLogger(__name__)


class Stopwords(BaseModel):
    """
    name: stopwords
    lowercase: true
    stopwords_fn:
    stopwords_list:
    stopwords_path:
    nltk_stopwords_lang:
    verbose: False
    """

    name: str = "stopwords"
    lowercase: bool = False
    stopwords_fn: Optional[Union[str, Callable[[str], bool]]] = None
    stopwords_list: Optional[List[str]] = None
    stopwords_path: Optional[str] = None
    nltk_stopwords_lang: Optional[str] = None
    verbose: bool = False

    _stopwords_fn: Optional[Callable] = None

    @model_validator(mode="after")
    def validate_stopwords(self) -> "Stopwords":
        if self.stopwords_path:
            self._stopwords_list = HyFI.load_wordlist(
                self.stopwords_path, lowercase=self.lowercase, verbose=self.verbose
            )
            if self.verbose:
                logger.info(
                    "Loaded %d stopwords from %s",
                    len(self._stopwords_list),
                    self.stopwords_path,
                )
        else:
            self._stopwords_list = []

        if callable(self.stopwords_fn):
            self._stopwords_fn = self.stopwords_fn
        elif isinstance(self.stopwords_fn, str):
            self._stopwords_fn = eval(self.stopwords_fn)

        if self.verbose:
            logger.info("Using custom stopwords function %s", self._stopwords_fn)

        if self.stopwords_list:
            if self.lowercase:
                self.stopwords_list = [w.lower() for w in self.stopwords_list]
            self._stopwords_list += self.stopwords_list

        if self.nltk_stopwords_lang:
            self._stopwords_list += self._load_nltk_stopwords(self.nltk_stopwords_lang)
        if self.verbose:
            logger.info("Loaded %d stopwords", len(self._stopwords_list))
        return self

    def __call__(self, word: str) -> bool:
        """Calling a stopwords instance like a function just calls the is_stopword method."""
        return self.is_stopword(word)

    def is_stopword(self, word: str) -> bool:
        """
        :type word: str
        :returns: bool
        """
        _word = word.lower() if self.lowercase else word
        if self._stopwords_fn:
            return self._stopwords_fn(_word) or (_word in self._stopwords_list)
        else:
            return _word in self._stopwords_list

    def _load_nltk_stopwords(self, language: str = "english") -> List[str]:
        """
        :type language: str
        :returns: list
        """
        import nltk
        from nltk.corpus import stopwords

        nltk.download("stopwords", quiet=True)
        if language in stopwords.fileids():
            logger.info("Loaded NLTK stopwords for %s", language)
            return stopwords.words(language)

        logger.warning("No NLTK stopwords for %s", language)
        return []

    def __iter__(self):
        return iter(self._stopwords_list)

    def __len__(self):
        return len(self._stopwords_list)

    def __contains__(self, word):
        return self.is_stopword(word)

    def __getitem__(self, word):
        return self.is_stopword(word)

    def __repr__(self):
        return f"<Stopwords {len(self._stopwords_list)} stopwords>"

    def __str__(self):
        return f"<Stopwords {len(self._stopwords_list)} stopwords>"

    def __bool__(self):
        return bool(self._stopwords_list)

    def __eq__(self, other):
        return self._stopwords_list == other._stopwords_list

    def __ne__(self, other):
        return self._stopwords_list != other._stopwords_list

    def __lt__(self, other):
        return len(self._stopwords_list) < len(other._stopwords_list)

    def __le__(self, other):
        return len(self._stopwords_list) <= len(other._stopwords_list)

    def __gt__(self, other):
        return len(self._stopwords_list) > len(other._stopwords_list)

    def __ge__(self, other):
        return len(self._stopwords_list) >= len(other._stopwords_list)

    def __add__(self, other):
        return Stopwords(stopwords_list=self._stopwords_list + other._stopwords_list)

    def __sub__(self, other):
        return Stopwords(
            stopwords_list=[
                w for w in self._stopwords_list if w not in other._stopwords_list
            ]
        )

    def __and__(self, other):
        return Stopwords(
            stopwords_list=[
                w for w in self._stopwords_list if w in other._stopwords_list
            ]
        )

    def __or__(self, other):
        return Stopwords(
            stopwords_list=self._stopwords_list
            + [w for w in other._stopwords_list if w not in self._stopwords_list]
        )

    def __xor__(self, other):
        return Stopwords(
            stopwords_list=[
                w for w in self._stopwords_list if w not in other._stopwords_list
            ]
            + [w for w in other._stopwords_list if w not in self._stopwords_list]
        )



import codecs
from typing import Callable, List, Optional, Tuple, Union

from pydantic import BaseModel

from lexikanon import HyFI
from lexikanon.normalizers import Normalizer
from lexikanon.stopwords import Stopwords

logger = HyFI.getLogger(__name__)


class Tokenizer(BaseModel):
    stopwords: Stopwords = Stopwords()
    normalizer: Normalizer = Normalizer()

    lowercase: bool = False
    strip_pos: bool = False
    postag_delim: str = "/"
    postag_length: Optional[int] = None
    include_whitespace_token: bool = True
    tokenize_each_word: bool = False
    sentence_separator: str = "\n"
    userdic_path: Optional[str] = None
    wordpieces_prefix: Optional[str] = "##"
    postags: Optional[List[str]] = None
    noun_postags: Optional[List[str]] = None
    punct_postags: List[str] = ["SF", "SP", "SSO", "SSC", "SY"]
    stop_postags: Optional[List[str]] = None
    verbose: bool = False

    @property
    def sentence_separator_unicode(self) -> str:
        return codecs.decode(self.sentence_separator, "unicode_escape")

    def __call__(self, text: str) -> List[str]:
        """Calling a tokenize config instance like a function just calls the tokenize method."""
        return self.tokenize(text)

    def tokenize(self, text: str) -> List[str]:
        if self.lowercase:
            text = text.lower()
        if self.normalizer and callable(self.normalizer):
            text = self.normalizer(text)
        if len(text) > 0:
            if self.tokenize_each_word:
                term_pos = []
                for word in text.split():
                    term_pos += self.tokenize_word(word)
            else:
                text = " ".join(text.split())
                term_pos = [self.to_token(token) for token in self.parse(text)]
        else:
            term_pos = []
        return term_pos

    def parse(self, text: str) -> List[Tuple[str, str]]:
        return [(token, "") for token in text.split()]

    def to_token(self, term_pos: Union[str, tuple]) -> str:
        if isinstance(term_pos, tuple):
            return _tuple_to_token(
                term_pos,
                self.strip_pos,
                self.postag_delim,
                self.postag_length,
            )
        return term_pos

    def tokenize_word(self, word: str) -> List[str]:
        tokens = self.parse(word)
        tokenized = []
        for i, token_pos in enumerate(tokens):
            term = self.to_token(token_pos)
            if i == 0 or not self.include_whitespace_token:
                tokenized.append(term)
                continue
            pos = token_pos[1]
            prv_pos = tokens[i - 1][1]
            if (
                self.wordpieces_prefix
                and pos not in self.punct_postags
                and prv_pos not in self.punct_postags
            ):
                term = f"{self.wordpieces_prefix}{term}"
            tokenized.append(term)
        return tokenized

    def pos(self, text: str) -> List[Tuple[str, str]]:
        return self.parse(text)

    def tokenize_article(
        self,
        article: Optional[str],
    ) -> List[List[str]]:
        if article is None:
            return []

        tokenized_article = []
        for sent in article.split(self.sentence_separator):
            sent = sent.strip()
            tokens = self.tokenize(sent)
            tokenized_article.append(tokens)
        return tokenized_article

    def nouns(self, text: str) -> List[str]:
        tokens = self.tokenize(text)
        return self.extract(tokens, nouns_only=True)

    def tokens(self, text: str) -> List[str]:
        tokens = self.tokenize(text)
        return self.extract(tokens, nouns_only=False)

    def morphs(self, text: str) -> List[str]:
        return self.tokens(text)

    def extract(
        self,
        text: Union[str, List[str]],
        nouns_only: bool = False,
        postags: Optional[List[str]] = None,
        stop_postags: Optional[List[str]] = None,
        strip_pos: Optional[bool] = None,
        postag_delim: Optional[str] = None,
        postag_length: Optional[int] = None,
    ) -> List[str]:
        if strip_pos is None:
            strip_pos = self.strip_pos
        if stop_postags is None:
            stop_postags = self.stop_postags
        if postags is None:
            postags = self.noun_postags if nouns_only else self.postags
        if postag_delim is None:
            postag_delim = self.postag_delim
        if postag_length is None:
            postag_length = self.postag_length

        return _extract_tokens(
            text,
            postags=postags,
            stop_postags=stop_postags,
            stopwords=self.stopwords,
            strip_pos=strip_pos,
            postag_delim=postag_delim,
            postag_length=postag_length,
        )

    def extract_article(
        self,
        article: Optional[str],
        nouns_only: bool = False,
    ) -> List[List[str]]:
        if article is None:
            return []

        tokens_article = []
        for sent in article.split(self.sentence_separator_unicode):
            sent = sent.strip()
            tokens = self.extract(sent, nouns_only=nouns_only)
            tokens_article.append(tokens)
        return tokens_article

    def extract_tokens(self, article: Optional[str]) -> List[List[str]]:
        return self.extract_article(article, nouns_only=False)

    def extract_nouns(self, article: Optional[str]) -> List[List[str]]:
        return self.extract_article(article, nouns_only=True)

    def filter_stopwords(
        self,
        text_or_tokens: Union[str, List[str], None],
    ) -> List[str]:
        if text_or_tokens is None:
            return []

        if isinstance(text_or_tokens, list):
            tokens = text_or_tokens
        else:
            tokens = self.tokenize(text_or_tokens)

        if self.stopwords and callable(self.stopwords):
            tokens = [token for token in tokens if not self.stopwords(token)]

        return tokens

    def filter_article_stopwords(self, article: Optional[str]) -> List[List[str]]:
        if article is None:
            return []

        return [
            self.filter_stopwords(sent)
            for sent in article.split(self.sentence_separator_unicode)
        ]


class SimpleTokenizer(Tokenizer):
    def parse(self, text: str) -> List[Tuple[str, str]]:
        return [(token, "") for token in text.split()]


def _match_tags(
    token: tuple,
    tags: List[str],
) -> bool:
    return any(token[1].startswith(tag) for tag in tags)


def _extract_tokens(
    tokenized_text: Union[str, List[str]],
    postags: Optional[List[str]] = None,
    stop_postags: Optional[List[str]] = None,
    stopwords: Optional[Callable[[str], bool]] = None,
    strip_pos: bool = True,
    postag_delim: str = "/",
    postag_length: Optional[int] = None,
    **kwargs,
) -> List[str]:
    if postags is None:
        postags = []
    if stop_postags is None:
        stop_postags = ["SP"]
    if isinstance(tokenized_text, str):
        tokens = tokenized_text.split()
    else:
        tokens = tokenized_text
    _token_pos_tuples = [
        _token_to_tuple(token, postag_delim=postag_delim, postag_length=postag_length)
        for token in tokens
    ]
    postags = [
        postag[:postag_length] if postag_length else postag
        for postag in postags
        if postag not in stop_postags
    ]
    stop_postags = [
        postag[:postag_length] if postag_length else postag for postag in stop_postags
    ]
    _tokens = []
    if postags:
        _tokens = [
            token_pos
            for token_pos in _token_pos_tuples
            if len(token_pos) == 1
            or (
                not _match_tags(token_pos, stop_postags)
                and _match_tags(token_pos, postags)
            )
        ]
    else:
        _tokens = [
            token_pos
            for token_pos in _token_pos_tuples
            if len(token_pos) == 1 or not _match_tags(token_pos, stop_postags)
        ]
    if stopwords is not None:
        _tokens = [token_pos for token_pos in _tokens if not stopwords(token_pos[0])]

    _tokens = [
        _tuple_to_token(
            token_pos,
            strip_pos=strip_pos,
            postag_delim=postag_delim,
            postag_length=postag_length,
        )
        for token_pos in _tokens
    ]
    return _tokens


def _tuple_to_token(
    token_pos: tuple,
    strip_pos: bool = True,
    postag_delim: str = "/",
    postag_length: Optional[int] = None,
) -> str:
    if strip_pos or len(token_pos) == 1:
        return token_pos[0]
    return (
        token_pos[0].strip()
        + postag_delim
        + (token_pos[1][:postag_length] if postag_length else token_pos[1])
    )


def _token_to_tuple(
    _token: Union[str, tuple],
    postag_delim: str = "/",
    postag_length: Optional[int] = None,
) -> Union[Tuple[str, str], Tuple[str]]:
    if isinstance(_token, str):
        token_pos = _token.split(postag_delim)
        if len(token_pos) == 2:
            return (
                token_pos[0],
                token_pos[1][:postag_length] if postag_length else token_pos[1],
            )
        return (token_pos[0],)
    return _token



from typing import List, Optional, Union

from datasets import Dataset

from lexikanon import HyFI

logger = HyFI.getLogger(__name__)


def tokenize_dataset(
    data: Dataset,
    tokenizer_config_name: str = "simple",
    num_workers: int = 1,
    batched: bool = True,
    batch_size: int = 1000,
    text_col: str = "bodyText",
    token_col: str = "tokenizedText",
    remove_columns: Optional[Union[List[str], str]] = None,
    load_from_cache_file: bool = True,
    num_heads: Optional[int] = 1,
    num_tails: Optional[int] = 1,
    verbose: bool = False,
) -> Dataset:
    def pos_tagging(batch):
        tokenizer = HyFI.instantiate_config(f"tokenizer={tokenizer_config_name}")
        batch_tokens = []
        for text in batch[text_col]:
            sentences = text.split("\n")
            tokens = []
            for sentence in sentences:
                tokens.extend(tokenizer(sentence))
            batch_tokens.append(tokens)
        return {token_col: batch_tokens}

    data = data.map(
        pos_tagging,
        num_proc=num_workers,
        batched=batched,
        batch_size=batch_size,
        remove_columns=remove_columns,
        load_from_cache_file=load_from_cache_file,
    )
    logger.info("POS tagging done. See column '%s'.", token_col)
    if verbose:
        if num_heads:
            num_heads = min(num_heads, len(data))
            print(data[:num_heads][token_col])
        if num_tails:
            num_tails = min(num_tails, len(data))
            print(data[-num_tails:][token_col])
    return data


def extract_tokens(
    data: Dataset,
    tokenizer_config_name: str = "simple",
    num_workers: int = 1,
    batched: bool = True,
    batch_size: int = 1000,
    token_col: str = "tokenizedText",
    extracted_col: str = "extractedTokens",
    nouns_only: bool = False,
    postags: Optional[List[str]] = None,
    stop_postags: Optional[List[str]] = None,
    strip_pos: Optional[bool] = None,
    postag_delim: Optional[str] = None,
    postag_length: Optional[int] = None,
    remove_columns: Optional[Union[List[str], str]] = None,
    load_from_cache_file: bool = True,
    num_heads: Optional[int] = 1,
    num_tails: Optional[int] = 1,
    verbose: bool = False,
) -> Dataset:
    def pos_tagging(batch):
        tokenizer = HyFI.instantiate_config(f"tokenizer={tokenizer_config_name}")
        batch_tokens = []
        for tokens in batch[token_col]:
            batch_tokens.append(
                tokenizer.extract(
                    tokens,
                    nouns_only=nouns_only,
                    postags=postags,
                    stop_postags=stop_postags,
                    strip_pos=strip_pos,
                    postag_delim=postag_delim,
                    postag_length=postag_length,
                )
            )
        return {extracted_col: batch_tokens}

    data = data.map(
        pos_tagging,
        num_proc=num_workers,
        batched=batched,
        batch_size=batch_size,
        remove_columns=remove_columns,
        load_from_cache_file=load_from_cache_file,
    )
    logger.info("Extracting tokens done, see column '%s'.", extracted_col)
    if verbose:
        if num_heads:
            num_heads = min(num_heads, len(data))
            print(data[:num_heads][extracted_col])
        if num_tails:
            num_tails = min(num_tails, len(data))
            print(data[-num_tails:][extracted_col])
    return data
